# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from tapl_language.pipeweaver import pipeweaver_language


def get_language():
    return pipeweaver_language.PipeweaverLanguage()
